from pathlib import Path
from typing import Dict


class Train:
    def __init__(
        self,
        config,
        output_dir: Path,
        variables: Dict,
        secrets,
    ):
        self._config = config
        self._output_dir = output_dir
        self._variables = variables
        self._secrets = secrets

    def train(self):
        # Write your training code here, populating generated artifacts in
        # self._output_dir
        pass



import logging
import os
from pathlib import Path

# TODO(pankaj) In desparate need of refactoring into separate library
try:
    from helpers.types import Action, Patch
except ModuleNotFoundError as exc:
    logging.debug(f"Importing helpers from truss core, caused by: {exc}")
    from truss.templates.control.control.helpers.types import Action, Patch


def apply_code_patch(
    relative_dir: Path,
    patch: Patch,
    logger: logging.Logger,
):
    logger.debug(f"Applying code patch {patch.to_dict()}")
    filepath: Path = relative_dir / patch.path
    action = patch.action

    if action in [Action.ADD, Action.UPDATE]:
        filepath.parent.mkdir(parents=True, exist_ok=True)
        action_log = "Adding" if action == Action.ADD else "Updating"
        logger.info(f"{action_log} file {filepath}")
        with filepath.open("w") as file:
            content = patch.content
            if content is None:
                raise ValueError(
                    "Invalid patch: content of a file update patch should not be None."
                )
            file.write(content)

    elif action == Action.REMOVE:
        if not filepath.exists():
            logger.warning(f"Could not delete file {filepath}: not found.")
        elif filepath.is_file():
            logger.info(f"Deleting file {filepath}")
            filepath.unlink()
            # attempt to recursively remove potentially empty directories, if applicable
            # os.removedirs raises OSError with errno 39 when this process encounters a non-empty dir
            try:
                os.removedirs(filepath.parent)
            except OSError as e:
                if e.errno == 39:  # Directory not empty
                    pass
                else:
                    raise
    else:
        raise ValueError(f"Unknown patch action {action}")



import tarfile
import tempfile
from pathlib import Path
from typing import IO, Any, Callable

from rich.progress import Progress


class ReadProgressIndicatorFileHandle:
    def __init__(
        self, file: IO[bytes], progress_callback: Callable[[int], Any]
    ) -> None:
        self._file = file
        self._progress_callback = progress_callback

    def read(self, size=-1):
        data = self._file.read(size)
        self._progress_callback(len(data))
        return data

    def __getattr__(self, attr):
        return getattr(self._file, attr)


def create_tar_with_progress_bar(source_dir: Path, delete=True):
    total_size = sum(f.stat().st_size for f in source_dir.glob("**/*") if f.is_file())

    # Keeping the .tgz suffix for backwards compatibility even though
    # this tar is uncompressed for upload
    temp_file = tempfile.NamedTemporaryFile(suffix=".tgz", delete=delete)
    with tarfile.open(temp_file.name, "w:") as tar:

        # Create a new progress bar
        progress = Progress()

        # Add a new task to the progress bar
        task_id = progress.add_task("[cyan]Compressing...", total=total_size)

        with progress:

            def file_read_progress_callback(bytes_read: int):
                # Update the progress bar
                progress.update(task_id, advance=bytes_read)

            for file_path in source_dir.glob("**/*"):
                if file_path.is_file():
                    arcname = str(file_path.relative_to(source_dir))
                    with file_path.open("rb") as file_obj:
                        file_obj_with_progress = ReadProgressIndicatorFileHandle(
                            file_obj, file_read_progress_callback
                        )
                        tarinfo = tar.gettarinfo(name=str(file_path), arcname=arcname)
                        tar.addfile(
                            tarinfo=tarinfo, fileobj=file_obj_with_progress  # type: ignore[arg-type]
                        )
    return temp_file



import logging
from pathlib import Path
from threading import Thread

import rich


class TrussFilesSyncer(Thread):
    """Daemon thread that watches for changes in the user's Truss and syncs to running service."""

    def __init__(
        self,
        watch_path: Path,
        remote,
    ) -> None:
        from truss.util.path import is_ignored, load_trussignore_patterns

        super().__init__(daemon=True)
        self._logger = logging.Logger(__name__)
        self.watch_path = watch_path
        self.remote = remote
        self.watch_filter = lambda _, path: not is_ignored(
            Path(path),
            load_trussignore_patterns(),
        )

    def run(self) -> None:
        """Watch for files in background and apply appropriate patches."""
        from watchfiles import watch

        # disable watchfiles logger
        logging.getLogger("watchfiles.main").disabled = True

        rich.print(f"🚰 Attempting to sync truss at '{self.watch_path}' with remote")
        self.remote.patch(self.watch_path, self._logger)

        rich.print(f"👀 Watching for changes to truss at '{self.watch_path}' ...")
        for _ in watch(
            self.watch_path, watch_filter=self.watch_filter, raise_interrupt=False
        ):
            self.remote.patch(self.watch_path, self._logger)



from pathlib import Path
from typing import Optional

from jinja2 import Environment, FileSystemLoader
from truss.constants import (
    REQUIREMENTS_TXT_FILENAME,
    SHARED_SERVING_AND_TRAINING_CODE_DIR,
    SHARED_SERVING_AND_TRAINING_CODE_DIR_NAME,
    SYSTEM_PACKAGES_TXT_FILENAME,
    TEMPLATES_DIR,
    TRAINING_DOCKERFILE_NAME,
    TRAINING_DOCKERFILE_TEMPLATE_NAME,
    TRAINING_JOB_WRAPPER_CODE_DIR,
    TRAINING_JOB_WRAPPER_CODE_DIR_NAME,
    TRAINING_REQUIREMENTS_TXT_FILENAME,
)
from truss.contexts.image_builder.image_builder import ImageBuilder
from truss.contexts.image_builder.util import (
    TRUSS_BASE_IMAGE_VERSION_TAG,
    file_is_not_empty,
    to_dotted_python_version,
    truss_base_image_name,
    truss_base_image_tag,
)
from truss.contexts.truss_context import TrussContext
from truss.truss_spec import TrussSpec
from truss.util.path import build_truss_target_directory, copy_file_path, copy_tree_path

BUILD_TRAINING_DIR_NAME = "training"


class TrainingImageBuilderContext(TrussContext):
    @staticmethod
    def run(truss_dir: Path):
        return TrainingImageBuilder(truss_dir)


class TrainingImageBuilder(ImageBuilder):
    def __init__(self, truss_dir: Path) -> None:
        self._truss_dir = truss_dir
        self._spec = TrussSpec(truss_dir)

    @property
    def default_tag(self):
        return f"{self._spec.model_framework_name}-train:latest"

    def prepare_image_build_dir(self, build_dir: Optional[Path] = None):
        """Prepare a directory for building the docker image from.

        Returns:
            docker command to build the docker image.
        """
        if build_dir is None:
            build_dir = build_truss_target_directory("train")

        copy_tree_path(self._spec.truss_dir, build_dir)

        if (self._spec.truss_dir / TRAINING_JOB_WRAPPER_CODE_DIR_NAME).exists():
            err_msg = (
                "Unable to build training image: training job "
                f"wrapper code directory name {TRAINING_JOB_WRAPPER_CODE_DIR_NAME}"
                " conflicts with an existing directory in the truss."
            )
            raise ValueError(err_msg)

        copy_tree_path(
            TRAINING_JOB_WRAPPER_CODE_DIR,
            build_dir / BUILD_TRAINING_DIR_NAME,
        )
        copy_tree_path(
            SHARED_SERVING_AND_TRAINING_CODE_DIR,
            build_dir
            / BUILD_TRAINING_DIR_NAME
            / SHARED_SERVING_AND_TRAINING_CODE_DIR_NAME,
        )

        copy_file_path(
            TEMPLATES_DIR / BUILD_TRAINING_DIR_NAME / REQUIREMENTS_TXT_FILENAME,
            build_dir / TRAINING_REQUIREMENTS_TXT_FILENAME,
        )
        with (build_dir / REQUIREMENTS_TXT_FILENAME).open("w") as req_file:
            req_file.write(self._spec.requirements_txt)

        with (build_dir / SYSTEM_PACKAGES_TXT_FILENAME).open("w") as req_file:
            req_file.write(self._spec.system_packages_txt)

        bundled_packages_dir_exists = (
            build_dir / self._spec.config.bundled_packages_dir
        ).exists()
        template_loader = FileSystemLoader(str(TEMPLATES_DIR))
        template_env = Environment(loader=template_loader)
        dockerfile_template = template_env.get_template(
            TRAINING_DOCKERFILE_TEMPLATE_NAME
        )
        should_install_system_requirements = file_is_not_empty(
            build_dir / SYSTEM_PACKAGES_TXT_FILENAME
        )
        should_install_requirements = file_is_not_empty(
            build_dir / REQUIREMENTS_TXT_FILENAME
        )
        config = self._spec.config
        python_version = to_dotted_python_version(config.python_version)
        if config.base_image:
            base_image_name_and_tag = config.base_image.image
        else:
            base_image_name = truss_base_image_name(job_type="training")
            tag = truss_base_image_tag(
                python_version=python_version,
                use_gpu=config.resources.use_gpu,
                version_tag=TRUSS_BASE_IMAGE_VERSION_TAG,
            )
            base_image_name_and_tag = f"{base_image_name}:{tag}"
        dockerfile_contents = dockerfile_template.render(
            base_image_name_and_tag=base_image_name_and_tag,
            config=self._spec.config,
            python_version=python_version,
            bundled_packages_dir_exists=bundled_packages_dir_exists,
            should_install_system_requirements=should_install_system_requirements,
            should_install_requirements=should_install_requirements,
        )
        docker_file_path = build_dir / TRAINING_DOCKERFILE_NAME
        with docker_file_path.open("w") as docker_file:
            docker_file.write(dockerfile_contents)



import logging
from pathlib import Path
from typing import List

from truss.templates.control.control.helpers.errors import UnsupportedPatch
from truss.templates.control.control.helpers.truss_patch.model_code_patch_applier import (
    apply_code_patch,
)
from truss.templates.control.control.helpers.truss_patch.requirement_name_identifier import (
    identify_requirement_name,
    reqs_by_name,
)
from truss.templates.control.control.helpers.truss_patch.system_packages import (
    system_packages_set,
)
from truss.templates.control.control.helpers.types import (
    Action,
    ConfigPatch,
    EnvVarPatch,
    ExternalDataPatch,
    ModelCodePatch,
    Patch,
    PythonRequirementPatch,
    SystemPackagePatch,
)
from truss.truss_config import TrussConfig


class TrussDirPatchApplier:
    """Applies patches to a truss directory.
    This should be compatible with ModelContainerPatchApplier.
    """

    def __init__(self, truss_dir: Path, logger: logging.Logger) -> None:
        self._truss_dir = truss_dir
        self._truss_config_path = self._truss_dir / "config.yaml"
        self._truss_config = TrussConfig.from_yaml(self._truss_config_path)
        self._logger = logger

    def __call__(self, patches: List[Patch]):
        # Apply model code patches immediately
        # Aggregate config patches and apply at end
        reqs = reqs_by_name(self._truss_config.requirements)
        pkgs = system_packages_set(self._truss_config.system_packages)
        new_config = self._truss_config
        for patch in patches:
            self._logger.debug(f"Applying patch {patch.to_dict()}")
            action = patch.body.action
            if isinstance(patch.body, ModelCodePatch):
                model_code_patch: ModelCodePatch = patch.body
                model_module_dir = self._truss_dir / self._truss_config.model_module_dir
                apply_code_patch(model_module_dir, model_code_patch, self._logger)
                continue
            if isinstance(patch.body, PythonRequirementPatch):
                py_req_patch: PythonRequirementPatch = patch.body
                req = py_req_patch.requirement
                req_name = identify_requirement_name(req)
                if action == Action.REMOVE:
                    del reqs[req_name]
                    continue
                if action == Action.ADD or Action.UPDATE:
                    reqs[req_name] = req
                continue
            if isinstance(patch.body, SystemPackagePatch):
                sys_pkg_patch: SystemPackagePatch = patch.body
                pkg = sys_pkg_patch.package
                if action == Action.REMOVE:
                    pkgs.remove(pkg)
                    continue
                if action == Action.ADD or Action.UPDATE:
                    pkgs.add(pkg)
                    continue
            # Each of EnvVarPatch and ExternalDataPatch can be expressed through an overwrite of the config,
            # handled below
            if isinstance(patch.body, EnvVarPatch):
                continue
            if isinstance(patch.body, ExternalDataPatch):
                continue
            if isinstance(patch.body, ConfigPatch):
                new_config = TrussConfig.from_dict(patch.body.config)
                continue
            raise UnsupportedPatch(f"Unknown patch type {patch.type}")

        new_config.write_to_yaml_file(self._truss_config_path)



import multiprocessing
import tempfile
import time
from pathlib import Path
from typing import Awaitable, Callable, List

import pytest
from truss.templates.server.common.termination_handler_middleware import (
    TerminationHandlerMiddleware,
)


async def noop(*args, **kwargs):
    return


@pytest.mark.integration
def test_termination_sequence_no_pending_requests(tmp_path):
    # Create middleware in separate process, on sending term signal to process,
    # it should print the right messages.
    def main_coro_gen(middleware: TerminationHandlerMiddleware):
        import asyncio

        async def main(*args, **kwargs):
            await middleware(1, call_next=noop)
            await asyncio.sleep(1)
            print("should not print due to termination")

        return main()

    _verify_term(main_coro_gen, ["stopped", "terminated"])


@pytest.mark.integration
def test_termination_sequence_with_pending_requests(tmp_path):
    def main_coro_gen(middleware: TerminationHandlerMiddleware):
        import asyncio

        async def main(*args, **kwargs):
            async def call_next(req):
                await asyncio.sleep(1.0)
                return "call_next_called"

            resp = await middleware(1, call_next=call_next)
            print(f"call_next response: {resp}")
            await asyncio.sleep(1)
            print("should not print due to termination")

        return main()

    _verify_term(
        main_coro_gen,
        [
            "stopped",
            "call_next response: call_next_called",
            "terminated",
        ],
    )


def _verify_term(
    main_coro_gen: Callable[[TerminationHandlerMiddleware], Awaitable],
    expected_lines: List[str],
):
    def run(stdout_capture_file_path):
        import asyncio
        import os
        import signal
        import sys

        sys.stdout = open(stdout_capture_file_path, "w")

        def term():
            print("terminated", flush=True)
            os.kill(os.getpid(), signal.SIGKILL)

        middleware = TerminationHandlerMiddleware(
            on_stop=lambda: print("stopped", flush=True),
            on_term=term,
            termination_delay_secs=0.1,
        )
        asyncio.run(main_coro_gen(middleware))

    stdout_capture_file = tempfile.NamedTemporaryFile()
    proc = multiprocessing.Process(target=run, args=(stdout_capture_file.name,))
    proc.start()
    time.sleep(1)
    proc.terminate()
    proc.join(timeout=6.0)
    with Path(stdout_capture_file.name).open() as file:
        lines = [line.strip() for line in file]

    assert lines == expected_lines



from pathlib import Path

import yaml
from triton_model_wrapper import TritonModelWrapper


class TritonPythonModel:
    def __init__(self):
        self._model_repository_path = None
        self._config_path = None
        self._config = None
        self._model_wrapper: TritonModelWrapper = None

    def _instantiate_model_wrapper(self, triton_config: dict) -> None:
        """
        Instantiates the model wrapper class as well as the user-defined model class.

        Args:
            triton_config (dict): Triton configuration dictionary. This contains information about the model repository
                path and the model version.

        Returns:
            None
        """
        self._model_repository_path = (
            Path(triton_config["model_repository"]) / triton_config["model_version"]
        )
        self._config_path = self._model_repository_path / "truss" / "config.yaml"
        with open(self._config_path, encoding="utf-8") as config_file:
            self._config = yaml.safe_load(config_file)
        self._model_wrapper: TritonModelWrapper = TritonModelWrapper(self._config)
        self._model_wrapper.instantiate()

    def initialize(self, triton_config: dict) -> None:
        """
        Instantiates the model wrapper class and loads the user's model. This function is called by Triton upon startup.

        Args:
            args (dict): Triton configuration dictionary. This contains information about the model repository path and
                the model version as well as all the information defined in the config.pbtxt.

        Returns:
            None
        """
        self._instantiate_model_wrapper(triton_config)
        self._model_wrapper.load()

    def execute(self, requests: list) -> list:
        """
        Executes the user's model on Triton InferenceRequest objects. This function is called by Triton upon inference.

        Args:
            requests (InferenceRequest): Triton InferenceRequest object. This contains the input data for the model.

        Returns:
            InferenceResponse: Triton InferenceResponse object. This contains the output data from the model.
        """
        return self._model_wrapper.predict(requests)



from InquirerPy import inquirer
from truss.truss_config import Build, ModelServer

REQUIRED_ARGUMENTS = {
    ModelServer.TGI: {
        "model_id": {
            "meta-llama/Llama-2-7b-chat-hf",
            "bigcode/starcoder",
            "google/flan-t5-xxl",
            "EleutherAI/gpt-neox-20b",
            "tiiuae/falcon-7b",
        },
        "endpoint": {"generate_stream", "generate"},
        # "dtype": {"bfloat16", "float16"},
    },
    ModelServer.VLLM: {
        "model": {
            "meta-llama/Llama-2-7b-chat-hf",
            "tiiuae/falcon-7b",
            # TODO(abu): Add more models
        },
        "endpoint": {"Completions", "ChatCompletions"},
    },
}


def select_server_backend(server_backend: ModelServer) -> Build:
    follow_up_questions = REQUIRED_ARGUMENTS.get(server_backend)
    args = {}
    if follow_up_questions:
        for q, opts in follow_up_questions.items():
            args[q] = inquirer.text(
                message=q, completer=dict({o: None for o in opts})
            ).execute()

    return Build(model_server=server_backend, arguments=args)


def ask_name() -> str:
    return inquirer.text(message="What's the name of your model?").execute()



import json
from typing import Any, List

import numpy as np
from utils.errors import UnsupportedTypeError


class Tensor:
    def as_numpy(self) -> np.ndarray:
        return  # type: ignore


class InferenceRequest:
    pass


class InferenceResponse:
    pass


def get_numpy_item(
    tensor: Tensor,
) -> str:
    """Returns the item of a numpy array as a Python type."""
    item = tensor.as_numpy().item()
    return item.decode("utf-8") if isinstance(item, bytes) else item


def convert_tensor_to_python_type(tensor: Tensor, dtype: str) -> Any:
    """Converts a Triton tensor to a Python type."""

    def _convert_tensor_to_dict(
        tensor: Tensor,
    ) -> dict:
        item = get_numpy_item(tensor)
        try:
            coerced_dict = json.loads(item)
        except json.decoder.JSONDecodeError:
            raise ValueError(
                f"Could not decode string to dict: {item}. Please ensure that the string is valid JSON."
            )
        return coerced_dict

    if dtype == "dict":
        return _convert_tensor_to_dict(tensor)
    elif dtype == "list":
        return tensor.as_numpy().flatten().tolist()
    elif dtype in ["str", "int", "float", "bool"]:
        return get_numpy_item(tensor)
    else:
        raise UnsupportedTypeError(f"Unsupported dtype: {dtype}")


def get_input_tensor_by_name(obj: InferenceRequest, name: str) -> Tensor:
    """Extracts a tensor from a Triton InferenceRequest by name."""
    from triton_python_backend_utils import get_input_tensor_by_name

    x = get_input_tensor_by_name(obj, name)
    if x is None:
        raise ValueError(f"Input tensor {name} not found in request.")
    return x


def create_tensor(name: str, value: Any) -> Tensor:
    """Creates a Triton tensor."""
    from triton_python_backend_utils import Tensor

    return Tensor(name, value)


def create_inference_response(
    objects: List[Tensor],
) -> InferenceResponse:
    """Creates a Triton InferenceResponse from a list of Triton tensors."""
    from triton_python_backend_utils import InferenceResponse

    return InferenceResponse(objects)
